import pandas as pd
from bigtree import nested_dict_to_tree, tree_to_nested_dict, print_tree, postorder_iter, preorder_iter, Node, shift_nodes
import yaml
import logging
from jinja2 import Environment, FileSystemLoader
import numpy as np
from memory import memory, get_register_bits_lists
from copy import deepcopy
import re
import os
from report import formatting
import xml.dom.minidom
import pathlib
import sys
import random


def attr_in_family(node, attr, value):
    for n in preorder_iter(node):
        if hasattr(n, attr):
            if value == getattr(n, attr):
                return True
    return False


def attr_in_children(node, attr, value):
    for n in preorder_iter(node.children):
        if hasattr(n, attr):
            if value == getattr(n, attr):
                return True
    return False


def get_full_name(node,direction):
    return node.path_name[1:].replace('/',f'{direction}')


def get_node_names(node,direction):
    names = {}
    if node.is_leaf:
        name = node.path_name[1:].replace('/', '_').lower()
        if node.width > 1:
            names['vector'] = f'std_logic_vector({node.width-1} downto 0)'
            names['zeroes'] = "(others => '0')"
            if hasattr(node,'stimulus'):
                names['stimulus'] = f'"{{value:0{node.width}b}}"'.format(value=node.stimulus)
        else:
            names['vector'] = f'std_logic'
            names['zeroes'] = "'0'"
            if hasattr(node, 'stimulus'):
                names['stimulus'] = f"'{node.stimulus}'"
        names['type_name'] = f'{name}_subtype'
        # full name for address decoder only
        names['full_name'] = node.path_name.replace(f'/{node.root.name}', f'{node.root.name}_{direction}').replace('/', '.').lower()
    else:
        name = node.path_name.replace(f'/{node.root.name}', f'{node.root.name}_{direction}').replace('/', '_').lower()
        names['type_name'] = f'{name}_record_type'

    # defining array and member names
    names['array_name'] = f'{name}_array_type'
    if hasattr(node,'length'):
        names['array'] = f'array ({node.length-1} downto 0) of {names["type_name"]}'
        names['member_name'] = names['array_name']
    else:
        names['array'] = ''
        names['member_name'] = names['type_name']

    return names


class wishlist(memory):
    def __init__(self, wishlist_file):
        self.tree = None
        self.wishlist_file = wishlist_file
        pathlib.Path("firmware").mkdir(parents=True, exist_ok=True)
        pathlib.Path("software").mkdir(parents=True, exist_ok=True)
        self.read_input_file()
        self.create_tree()
        self.computing_width()
        self.set_jinja_environment()
        self.generate_vhdl_file(template="vhdl_package.jinja2", suffix='pkg')
        # starting memory object
        super().__init__(start=self.tree.address, end=self.tree.address + self.tree.address_size - 1,
                         width=self.tree.address_width, increment=self.tree.address_increment)
        self.flattening()
        # Assigning stimulus for instantiation example and hardware validation
        self.generating_stimulus()
        #print_tree(self.tree, attr_list=['address', 'mask', 'width', 'length', 'permission', 'description', 'stimulus'])
        # Allocation
        self.address_decoder_list = []
        for node in self.register_nodes_iter():
            self.allocate(node)
        # Writing back-annotated yam file
        self.write_yaml_file(tree_to_nested_dict(self.tree,all_attrs=True),f"{self.wishlist_dict['firmware_path']}/{self.wishlist_dict['name'].lower()}_backannotated.yaml")
        print_tree(self.tree,all_attrs=True)
        # Generating software description file
        self.generate_uhal_file()
        # Generating address decoder tables and VHDL code
        self.address_decoder = pd.concat(self.address_decoder_list)
        self.address_decoder.to_html(f"{self.wishlist_dict['firmware_path']}/{self.wishlist_dict['name'].lower()}_address_decoder_verbose.htm")
        self.generate_vhdl_address_decoder_file()
        self.generate_vhdl_file(template="vhdl_instantiation.jinja2", suffix='instantiation')
        self.generate_vhdl_file(template="vhdl_axilite.jinja2", suffix='axilite')
        self.generate_vhdl_file(template="vhdl_accumulators.jinja2", suffix='accumulators')
        # Dropping unused address offsets
        self.space = self.space.dropna(how='all')
        self.space_style = self.space_style.loc[self.space.index,:]
        # Formatting space and its style
        self.space, self.space_style = formatting(self.space, self.space_style, self.wishlist_dict)
        # Creating styler object from dataframe with values and dataframe with CSS string
        self.update_style()
        # Rendering styler object
        self.space_styled.hide(axis="index").to_html(buf=f"{self.wishlist_dict['software_path']}/{self.wishlist_dict['name'].lower()}_address_space.htm")
        #self.space.fillna('').to_latex(buf=f"{self.wishlist_dict['firmware_path']}/{self.wishlist_dict['name'].lower()}_address_space.latex",label=self.wishlist_dict['name'].lower(), index=False, longtable=True)

    def read_input_file(self):
        with open(self.wishlist_file, "r") as stream:
            try:
                self.wishlist_dict = yaml.safe_load(stream)
            except yaml.YAMLError:
                logging.exception(f'Error while reading {self.wishlist_file}.')
            # Making sure address is read as HexInt
            self.wishlist_dict['address'] = HexInt(self.wishlist_dict['address'])
            self.wishlist_dict['address_size'] = HexInt(self.wishlist_dict['address_size'])

    def write_yaml_file(self, dictionary, filename):
        with open(filename, 'w') as file:
            yaml.add_representer(HexInt, representer)
            yaml.dump(dictionary, file, sort_keys=False)

    def create_tree(self):
        self.tree = nested_dict_to_tree(self.wishlist_dict)
        for node in preorder_iter(self.tree):
            if re.match('^(?!.*__)[a-zA-Z][\w]*[^_]$', node.name) is None:
                raise ValueError(f'Node name {node.name} in {node.path_name} is not a valid VHDL indentifier. Please change the name.')

    def flattening(self):
        # Flattening tree
        finished = False
        while not finished:
            finished = True
            for node in preorder_iter(self.tree):
                if hasattr(node,'length'):
                    for i in range(node.length):
                        attributes = dict(node.describe(exclude_attributes=["name", 'address', 'parent', 'children', 'description', 'length'],exclude_prefix="_"))
                        children = deepcopy(node.children)
                        if hasattr(node,'description'):
                            attributes['description'] = f'Instance {i}; {node.description}'
                        if hasattr(node, 'address'):
                            if i == 0:
                                attributes['address'] = node.address
                        # if hasattr(node, 'mask'):
                        #     attributes['mask'] = node.mask
                        # if hasattr(node, 'width'):
                        #     attributes['width'] = node.width
                        # if hasattr(node, 'permission'):
                        #     attributes['permission'] = node.permission
                        a = Node(f'{node.name}({i})', parent=node.parent, children=list(children), **attributes)
                    shift_nodes(self.tree, [node.path_name[1:]], [None])
                    finished = False
                    break
                    #print_tree(self.tree, attr_list=['address', 'mask', 'width', 'length', 'permission'])

    def generating_stimulus(self):
        # Generating random number using path_name as seed, so the number is unchanged in different runs
        for node in self.register_nodes_iter():
            random.seed(node.path_name)
            node.stimulus = random.randint(0, 2 ** node.width - 1)
    def register_nodes_iter(self):
        return preorder_iter(self.tree, filter_condition=lambda node: node.is_leaf)

    def hierarchical_nodes_iter(self):
        return postorder_iter(self.tree, filter_condition=lambda node: not node.is_leaf)

    def computing_width(self):
        for node in self.register_nodes_iter():
            if hasattr(node, 'mask'):
                mask_str = f'{node.mask:b}'
                diff = np.diff([int(s) for s in mask_str])
                if len(diff[diff != 0]) > 1:
                    raise ValueError(
                        f'The specified mask=0b{mask_str} for node {node.path_name} is asserted high in a non-contiguous bit interval\n.'
                        f'Currently, whishlist does not suport this use case, i.e. all the mask bits asserted high have to be grouped together.'
                    )
                width = int(mask_str.count('1'))
                if hasattr(node, 'width'):
                    if node.width != width:
                        raise ValueError(f'Both mask and width have been defined for {node.path_name}. However the width computed from the mask={width} is different than the specified width={node.width}.')
                else:
                    node.width = width
            elif not hasattr(node, 'width'):
                raise ValueError(
                    f'Node {node.path_name} has no width or mask specified. One of them is required, and if both are defined,'
                    f'the mask value gets priority over the width as it enforces the bit position in the address offset.\n'
                    f'Nevertheless, the provided width is still checked against the provided mask for consistency checking.'
                )


    def allocate(self, node):
        if hasattr(node,'address'):
            smart = False
            if hasattr(node, 'mask'):
                msb = self.tree.address_width-1-f'{{mask:0{self.tree.address_width}b}}'.format(mask=node.mask).find('1')
                self.set_bit_cursor(msb)
            else:
                if self.address != node.address:
                    self.set_bit_cursor(self.tree.address_width-1)
            self.set_address_cursor(node.address)
        else:
            smart = True
        # Allocating
        address_list, address_bits_lists = self.allocate_from_width(width=node.width, name=node.path_name, permission=node.permission, smart=smart)
        # creating respective dataframe and appending to the address_decoder_list
        direction = {
            'rw': 'control',
            'r': 'status',
        }
        self.address_decoder_list.append(pd.DataFrame({
            'name': [node.path_name]*len(address_list),
            'permission': [node.permission]*len(address_list),
            'address': address_list,
            'address_bits_lists': address_bits_lists,
            'register_bits_lists': get_register_bits_lists(address_list, address_bits_lists, node.width),
            'vhdl_member_name': [get_node_names(node, direction=direction[node.permission])['full_name']]*len(address_list),
            'node_width' : [node.width]*len(address_list)
        }))
        # Back-annottating address and mask
        node.address = [HexInt(addr) for addr in address_list]
        node.mask = [HexInt(self.bit_list_to_mask(bits_list)) for bits_list in address_bits_lists]


    def bit_list_to_mask(self, bit_list):
        # it works even for non-contiguous bit lists
        mask_list = np.zeros(max(bit_list)+1,dtype=int)
        mask_list[bit_list] = 1
        mask = sum([x << i for i, x in enumerate(mask_list)])
        return int(mask)

    def get_address_string(self, address):
        return f'{{address:0{np.ceil(self.tree.address_width/4).astype(int)}X}}'.format(address=address)
    def get_vhdl_bit_string(self, bits, side, node_width):
        if node_width == 1 and side == 'signal':
            return ''
        elif len(bits) == 1:
            return f'({bits[0]})'
        else:
            return f'({bits[0]} downto {bits[-1]})'

    # def get_signal_name(self, name, permission):
    #     direction_dict = {'r': 'status_int', 'rw': 'control_int'}
    #     return name.replace(f'/{self.tree.name}/', f'{self.tree.name}_{direction_dict[permission]}/').replace('/', '.').lower()



    def set_jinja_environment(self):
        self.environment = Environment(loader=FileSystemLoader("templates/"))
        self.environment.globals['attr_in_children'] = attr_in_children
        self.environment.globals['attr_in_family'] = attr_in_family
        self.environment.globals['get_full_name'] = get_full_name
        self.environment.globals['get_node_names'] = get_node_names

    def generate_vhdl_file(self, template, suffix):
        template = self.environment.get_template(template)
        filename = f"{self.wishlist_dict['firmware_path']}/{self.wishlist_dict['name'].lower()}_{suffix}.vhd"
        content = template.render(self.wishlist_dict,
                                  registers=list(self.register_nodes_iter()),
                                  hierarchies=list(self.hierarchical_nodes_iter()),
                                  )
        with open(filename, mode="w") as message:
            message.write(content)

    def generate_vhdl_address_decoder_file(self):
        template = self.environment.get_template("vhdl_address_decoder.jinja2")
        filename = f"{self.wishlist_dict['firmware_path']}/{self.wishlist_dict['name'].lower()}_address_decoder.vhd"
        content = template.render(self.wishlist_dict,
                                  registers=list(self.register_nodes_iter()),
                                  hierarchies=list(self.hierarchical_nodes_iter()),
                                  address_decoder=self.address_decoder,
                                  np=np,
                                  get_address_string=self.get_address_string,
                                  get_vhdl_bit_string=self.get_vhdl_bit_string,
                                  # get_signal_name = self.get_signal_name,

                                  )
        with open(filename, mode="w") as message:
            message.write(content)

    def generate_uhal_file(self):
        # Masking sure there are no words larger than 32 bits
        for node in self.register_nodes_iter():
            if node.width > 32:
                print(f'Omitting node {node.path_name} in XML output because uHAL does not support width higher than 32.')
        # Adding first addr to parent node recursively
        for node in postorder_iter(self.tree, filter_condition=lambda node: not node.is_root):
            node.parent.address = node.parent.children[0].address
        # Now converting absolute to relative addresses (requirement from uHAL), making sure top remains with addr 0x0
        for node in postorder_iter(self.tree, filter_condition=lambda node: node.path_name.count('/') > 2):
            node.address = [node.address[0]-node.parent.address[0]]
        # Rendering XML file
        template = self.environment.get_template("xml_uhal.jinja2")
        content = template.render(tree=self.tree)
        filename = f"{self.wishlist_dict['software_path']}/{self.wishlist_dict['name'].lower()}_uhal.xml"
        # Trying to beautify, if fails write rendered version
        with open(filename, mode="w") as message:
            try:
                dom = xml_beautify(content)
            except:
                print('XML ERROR, please check output XML')
                dom = content
            message.write(dom)



# HexInt class and respective representer
class HexInt(int): pass


def representer(dumper, data):
    n_bits = np.ceil(np.log2(data))
    if n_bits <= 32:
        n_hex = 8
    else:
        n_hex = int(np.ceil(n_bits/4))
    return yaml.ScalarNode('tag:yaml.org,2002:int', f'0x{{data:0{n_hex:d}X}}'.format(data=data))


def xml_beautify(content):
    dom = xml.dom.minidom.parseString(content)
    dom_string = dom.toprettyxml()
    return os.linesep.join([s for s in dom_string.splitlines() if s.strip()])


if __name__ == '__main__':
    obj = wishlist(sys.argv[1])
    #for path in ['firmware/l1calogfex_backannotated.yaml', 'firmware/l1calogfex_pkg.vhd', 'firmware/l1calogfex_address_decoder.vhd', 'firmware/l1calogfex_instantiation.vhd', 'examples/L1CaloGfex.yaml']:
        #os.system(f'scp {path} aiatlas-fw-03.cern.ch:git/register/zfpga/zfpga_top/source/')


import os
from bigtree import preorder_iter, find_name
from wishlist_axi_node import wishlist_axi_node
from utils import get_logger, log_tree, read_tree
import logging
import random
import socket
import sys
import time
import pandas as pd
from pathlib import Path
from datetime import datetime
from pytz import timezone


class wishlist_robot(object):
    def __init__(self, yaml_file=None, log_level=logging.INFO):
        self.logger = get_logger('Wishlist Robot', log_level)
        self.logger.info(f'Starting robot in {socket.gethostname()} using the register tree shown below loaded from {os.getenv("BACKANNOTATED_YAML")}')
        self.tree = read_tree(yaml_file,wishlist_axi_node)
        log_tree(self.tree, self.logger)

    def stress_test(self, nodes=None, N=1000, test_only_rw=False):
        start_time = time.time()
        # Testing the entire tree if nodes is not defined
        if nodes is None:
            nodes = self.tree
        # Making sure only leaves and rw registers are tested
        if test_only_rw:
            nodes = [n for n in nodes if (n.permission == 'rw' and n.is_leaf)]
        for node in nodes:
            self.logger.info(f'Running stress test with node {node.path_name}')
        # Testing all nodes N times
        for i in range(1,N+1):
            # Shuffling nodes (not in-place) order before writing, in-place (random.shuffle) shuffling was causing problems during writing
            nodes = random.sample(nodes, len(nodes))
            for node in nodes:
                # Generating stimulus and writing if register permission is rw
                if node.permission == 'rw':
                    node.stimulus = random.randint(0, 2 ** node.width - 1)
                    # Writing stimulus
                    node.write(node.stimulus)
            # Shuffling nodes order before reading
            nodes = random.sample(nodes, len(nodes))
            for node in nodes:
                # Reading node
                value = node.read()
                # Checking read value against stimulus for errors
                if value != node.stimulus:
                    node.logger.critical(f'node check error at iteration {i}: {value} (0x{value:08x}), expected: {node.stimulus}')
                    sys.exit()
                node.logger.debug(f'Read value is {value} (0x{value:08x})and stimulus is  {node.stimulus}')
            # Logging status every 100 iterations
            if not i % 1000:
                self.logger.info(f'Stress test iteration {i} out of {N} elapsed {time.time() - start_time} seconds')
        self.logger.info(f'Stress test with {N} iteration finished in {time.time() - start_time} seconds without errors')

    def launch_online_monitoring(self, monitored_nodes, clear_load_node, time_reference_node, timer_node, display=True, save=True, base_path='/software/tmp'):
        Path(f'{base_path}').mkdir(parents=True, exist_ok=True)
        accumulators_df = pd.DataFrame({'Value': [0] * len(monitored_nodes)},
                                       index=[node.name for node in monitored_nodes])
        i = 0
        save_df_list = []
        while True:
            try:
                # Waiting for reference to reach desired time period
                while timer_node.read() < 500e6:
                    time.sleep(0.1)
                clear_load_node.write(1)
                clear_load_node.write(0)
                time_reference = time_reference_node.read()
                if time_reference == (1 << time_reference_node.width) - 1:
                    self.logger.warning(f'Online monitoring iteration {i} has been discarded because time reference value is saturated.')
                else:
                    now = datetime.now(timezone('Europe/Paris'))
                    save_dict = {'time_reference': time_reference}
                    for node in monitored_nodes:
                        value = node.read()
                        rate = node.convert(value=value, reference=time_reference, parameter="conversion",)
                        if display:
                            representation = node.convert(value=value, rate=rate, parameter="representation")
                            accumulators_df.loc[
                                node.name, 'Value'] = f'{representation}'
                        if save:
                            save_dict[node.name] = rate

                    if display:
                        # Generating string before clearing the screen to avoid flickering
                        #df_str = accumulators_df.loc[['clk' in i for i in accumulators_df.index],:].to_string(col_space=30)
                        df_str = accumulators_df.to_string(col_space=30)
                    if save:
                        save_df = pd.DataFrame(save_dict, index=[now])
                        save_df_list.append(save_df)
                        if not i % 20:
                            save_dfs = pd.concat(save_df_list)
                            now_str = f'{now}'.replace(' ','_')
                            filename = f'{base_path}/online_monitoring_data_{now_str}.pickle'
                            save_dfs.to_pickle(filename)
                            self.logger.info(f'Iteration {i} - saved file {filename} ')
                            save_df_list = []
                    if display:
                        os.system('clear')
                        self.logger.info(f'Iteration {i} - {now} - refresh time: {time_reference * 10e-9} seconds ')
                        print(df_str)
                i += 1
            except KeyboardInterrupt:
                sys.exit()


if __name__ == '__main__':
    robot = wishlist_robot(yaml_file='../firmware/l1calogfex_backannotated.yaml', log_level=logging.INFO)
    nodes = list(preorder_iter(robot.tree, filter_condition=lambda node: node.is_leaf))
    for node in nodes:
        robot.logger.info(f'{node.path_name}: 0x{node.read():x}')
    nodes = list(preorder_iter(robot.tree, filter_condition=lambda node: node.is_leaf and 'test_' in node.name))
    robot.stress_test(nodes, N=10)
    robot.logger.info(f"Init status {find_name(robot.tree,'INIT_STAT').read():08x}")
    monitored_nodes = list(
        preorder_iter(robot.tree, filter_condition=lambda node: (node.is_leaf and node.permission == 'r') and (
                    '/monitoring/' in node.path_name or 'INIT_STAT' in node.name)))
    clear_load_node = find_name(robot.tree, 'clear_load')
    time_reference_node = find_name(robot.tree, 'ps_sys_clk')
    timer_node = find_name(robot.tree, 'ps_sys_clk_no_shadow')
    robot.launch_online_monitoring(monitored_nodes=monitored_nodes,
                                   clear_load_node=clear_load_node,
                                   time_reference_node=time_reference_node,
                                   timer_node=timer_node,
                                   display=False)









import pandas as pd
import numpy as np


def str2int(v):
    if isinstance(v, list):
        return [str2int(i) for i in v]
    elif isinstance(v, str):
        if v.startswith('0x'):
            return int(v, 16)
        else:
            return int(v, 10)
    else:
        raise Exception('The input needs to be a string or a list of string')


def str_range2list(string):
    return bounds2list(str2int(string.split(':')))


def bounds2list(bounds):
    if len(bounds != 2):
        raise Exception('Each range string should contain only two integer values')
    return list(inclusive_range(*bounds, increment=1))


def inclusive_range(start, end, increment=1):
    if start <= end:
        return range(start, end + 1, increment)
    else:
        return range(start, end - 1, increment)


def check_list(list_of, dtype=int):
    if isinstance(list_of, list):
        if all(isinstance(elem, dtype) for elem in list_of):
            return True
        else:
            raise Exception(f'All elements of the list should be of type {dtype.__name__}')
    else:
        raise Exception('The input should be a list')


def check_list_of_list(list_of_lists, type):
    return all(check_list(elem, type) for elem in list_of_lists)


def get_register_bits_lists(address_list, address_bits_lists, width):
    register_current_bit = width - 1
    register_bits_lists = []
    # Iterating through the list of addresses and list of list of address_bits
    for address, address_bits in zip(address_list, address_bits_lists):
        register_bits = []
        # assigning name of the owner for each bit
        for bit in address_bits:
            # keep tracking of the register bit
            register_bits.append(register_current_bit)
            # decrementing current bit
            register_current_bit -= 1
        # keep track of the list of register_bits
        register_bits_lists.append(register_bits)
    return register_bits_lists



class memory:
    def __init__(self, start=0, end=2 ** 32 - 1, width=32, increment=1):
        self.start = start
        self.end = end
        self.width = width
        self.increment = increment
        self.address = self.start
        self.set_address_cursor(self.start)
        self.bit = self.width-1
        self.space = pd.DataFrame(None, index=inclusive_range(self.start, self.end, self.increment),
                                  columns=inclusive_range(self.width - 1, 0, -1))
        self.color = {
            'unallocated' : 'Gainsboro',
            'smart_allocated_rw' : 'Plum',
            'hard_allocated_rw' : 'DeepSkyBlue',
            'smart_allocated_r': 'Gold',
            'hard_allocated_r': 'LightGreen',
        }

        self.space_style = pd.DataFrame(self.get_css_style(allocated=False), index=self.space.index, columns=self.space.columns)
        self.space_styled = pd.DataFrame(None, index=inclusive_range(self.start, self.end, self.increment),
                                  columns=inclusive_range(self.width - 1, 0, -1))

    def get_css_style(self, allocated=True, **kwargs):
        if allocated:
            smart_string = ('hard','smart')[kwargs['smart']]
            color_name = f'{smart_string}_allocated_{kwargs["permission"]}'
        else:
            color_name = 'unallocated'
        return 'border: 1px solid black; background-color: {c:s}'.format(c=self.color[color_name])

    def update_style(self):
        # This is a function that returns the dataframe containing the CSS strings.
        # For some reason it cant be a member of the class otherwise a recursion problem arises
        # and python reaches the maximum of 1000-depth recursion
        get_space_style = lambda df: self.space_style
        self.space_styled = self.space.fillna("").style.apply(get_space_style, axis=None)

    def set_address_cursor(self, address):
        if self.start <= address <= self.end:
            if not (address - self.start) % self.increment:
                self.address = address
            else:
                raise Exception(
                    "The address offset with respect to the start address is not a multiple of the increment value")
        else:
            raise Exception("The address is not within the start and end values")

    def set_bit_cursor(self, bit):
        if 0 <= bit <= self.width - 1:
            self.bit = bit
        else:
            raise Exception("The bit is not in the range 0 <= bit <= width-1")

    def print_debug_space(self):
        print('Debugging information: \n',
              self.space.loc[self.address - 2 * self.increment:self.address + 2 * self.increment, :])

    def _is_row_available(self, address, bits):
        return self.space.loc[address, bits].dropna().empty

    def _is_available(self, address_list, bits_list):
        return all([self._is_row_available(address, bits) for address, bits in zip(address_list, bits_list)])

    def is_available(self, address_list, bits_list):
        if check_list(address_list, int) and all([address in self.space.index for address in address_list]):
            if check_list_of_list(bits_list, int) and all(
                    [item in self.space.columns for sublist in bits_list for item in sublist]):
                if len(address_list) == len(bits_list):
                    return self._is_available(address_list, bits_list)
                else:
                    raise Exception('The list of address should have the same length as the list of bit list')
            else:
                raise Exception(
                    'All elements of the bits list should be integer and in the range 0 <= bit <= width-1')
        else:
            raise Exception(
                'All elements of the address list should be integer and in the range within the start and end values. Check if the requested width is too large.')


    def allocate_from_width(self, width, name=None, permission=None, smart=True):
        if name is None: name = '__allocated_without_name__'
        if permission is None: permission = '__allocated_without_permission__'
        # checking if requested width fits in current address offset
        if self.bit >= width - 1:
            address_bits_lists = [list(range(self.bit, self.bit - width, -1))]
        elif smart:
            # By convention, if a node cant be fully allocated in the current address, it attempts to allocate it on
            # the next address offset always starting from the MSB. But the address offset is only set in the while loop
            self.set_bit_cursor(self.width-1)
            remainder = width % self.width
            address_bits_lists = [list(inclusive_range(self.width - 1, 0, -1))] * (width // self.width) + \
                         ([list(range(self.width - 1, self.width - remainder - 1, -1))],[])[remainder==0]
        else:
            raise Exception(f'Allocation will not check if the required {width} bits are already in use in the memory-mapped space because the smart mode is off and there are not enough bits for any address offset using address width={self.width} and bit cursor={self.bit} to accomodate the required memory space.')

       # Computing bits lists required starting from the MSB
        while True:
            # List of address offsets to be requested
            # Here, I am just using sequential address values, one day one might evaluate using a different strategy
            # to minimize address decoder combinational logic
            address_list = list(range(self.address, self.address + np.ceil(width/self.width).astype(int)*self.increment, self.increment))
            # Checking if the requested addresses and bits lists are available
            if self.is_available(address_list, address_bits_lists):
                # Filling the memory space with the owner of each memory space bit
                final_address, final_address_bit = self._assign_owner(address_list, address_bits_lists, name, width, permission, smart)
                # moving the cursor to the end of the allocated space and incrementing one bit
                self.set_address_cursor(final_address)
                self.set_bit_cursor(final_address_bit)
                self.bit_increment()
                return address_list, address_bits_lists
            elif smart:
                # If smart mode is on, keep trying to allocate until the end address is reached
                if self.address <= self.end:
                    self.address_increment()
                else:
                    raise Exception('Smart allocation is unable to find the requested memory space and reached the maximum address value')
            else:
                self.print_debug_space()
                raise Exception(f'Allocation is unable to find the requested memory space for {name} with ({width} bits) with address={self.address} and bit={self.bit} cursors with smart mode off. The entire or a subset of the requested memory space is already in use. It wont keep trying because smart mode is off.')

    def _assign_owner(self, address_list, address_bits_lists, name, width, permission, smart):
        # Iterating through the list of addresses and list of list of address_bits
        # address_bits represent the bits in the AXI address signal, i.e. right side of the address decoder assignment for write transactions
        # register_bits represent the bits in the register signal, i.e. left side of the address decoder assignment for write transactions
        register_bits_lists = get_register_bits_lists(address_list, address_bits_lists, width)
        for address, address_bits, register_bits in zip(address_list, address_bits_lists, register_bits_lists):
            # assigning name of the owner for each bit
            for address_bit, register_bit in zip(address_bits, register_bits):
                # If width > 1, add bit index between brackets
                if width > 1:
                    self.space.loc[address, address_bit] = f'{name}({register_bit})'
                # otherwise, just the name
                else:
                    self.space.loc[address, address_bit] = f'{name}'
                # adding a CSS string to each cell
                self.space_style.loc[address, address_bit] = self.get_css_style(permission=permission, smart=smart)
        return address, address_bit


    def address_increment(self):
        self.set_address_cursor(self.address + self.increment)

    def bit_increment(self):
        # If current bit cursor is greater than 0, just decrement the bit cursor
        if self.bit > 0:
            self.set_bit_cursor(self.bit-1)
        # Otherwise, increment the address offset and move the bit cursor to the MSB
        else:
            self.address_increment()
            self.set_bit_cursor(self.width - 1)




if __name__ == '__main__':
    obj = memory(start=0, end=2 ** 7 - 1, width=32, increment=4)
    obj.space.loc[4,6] = 'hi'
    print(obj.allocate_from_width(15, name='hi', permission='rw', smart=False))
    obj.address_increment()
    print(obj.allocate_from_width(5, name='hi2', permission='r', smart=False))
    print(obj.allocate_from_width(200, name='hi3', permission='r' ))
    print(obj.allocate_from_width(5, name='hi4', permission='rw', smart=True))
    obj.update_style()


    print()



from Pandas2Plot import *
from multiprocessing import Pool
from datetime import datetime,timedelta


# Function outside the class to allow multithreading
def create_Pandas2Plot(larc, plot_threads, plot_vector_items, partition, start, stop, run, load_pickle, save_pickle,
                       wanted_variable_regexes):
    pandas2plot = Pandas2Plot(larc=larc,
                plot_threads=plot_threads,
                plot_vector_items=plot_vector_items,
                partition=partition,
                start=start,
                stop=stop,
                run=run,
                load_pickle=load_pickle,
                save_pickle=save_pickle,
                wanted_variable_regexes=wanted_variable_regexes,
                )
    pandas2plot.action()

class MultiPandas2Plot(Pandas2Plot):
    def __init__(self,larcs,larc_threads,**kwargs):
        super().__init__(**kwargs)
        self.larcs = larcs
        self.larc_threads = larc_threads
        self.menu_fmt = '<html>\n<table style="width:100%">\n<tr>\n<td>\n<a href="../../index.html" target="_top"><img src="../larisplot_small.png" width="40%" height="40%"></a></td>\n{cols:s}\n</tr>\n</table>\n</html>'
        self.menu_col_fmt = '\n<td>{col:s}</td>'
        self.menu_row_fmt = '<a href = "{link:s}" target = "{target:s}">{name:s}</a><br>\n'

    def action(self):
        # dont generate larc menu and keep the one generated under www
        #self.write_menu_table(f'../www/{self.run}/larc_menu.html', self.larcs, target='main_top',
                              #link_fmt='{item:s}/index.html', max_row=5)
        os.system(f'cd ../www/{self.run}/; ln -fs ../perrun_index.html index.html')
        os.system(f'cd ../www/{self.run}/; ln -fs ../larc_menu.html larc_menu.html')

        cre_Pandas2Plot = partial(create_Pandas2Plot, plot_threads=self.plot_threads,
                                  plot_vector_items=self.plot_vector_items, partition=self.partition,
                                  start=self.start, stop=self.stop, run=self.run, load_pickle=self.load_pickle,
                                  save_pickle=self.save_pickle, wanted_variable_regexes=None)
        # Multi and single threads calls
        if self.larc_threads == 1:
            for lar in self.larcs:
                cre_Pandas2Plot(lar)
        elif 1 <= self.larc_threads <= 5:
            if self.plot_threads == 1:
                with Pool(self.larc_threads) as p:
                    p.map(cre_Pandas2Plot, self.larcs)
            else:
                self.lgr.fatal('Plot and LArC multithreading can not coexist because daemonic processes are not allowed to have children. If many IS variables are wanted, opt for plot multithreading by setting plot threads to None and larc threads to 1.')
        else:
            self.lgr.fatal('LArC Threads has to be from 1 to 5. The PBeast server drops requests if more than 5 queries are submitted simultaneously.')

    def write_menu_table(self, filepath, items, target, link_fmt='{item:s}.html}', max_row=2):
        cols = ''
        col = ''
        for i, item in enumerate(items):
            link = link_fmt.format(item=item)
            col += self.menu_row_fmt.format(link=link, target=target, name=item)
            if i % max_row == (max_row - 1):
                cols += self.menu_col_fmt.format(col=col)
                col = ''
        cols += self.menu_col_fmt.format(col=col)
        self.write_str(self.menu_fmt.format(cols=cols), filepath)




if __name__ == '__main__':

    larcs = ['gfex-production-stf']

    now = datetime.now()
    stop = now.strftime("%Y-%m-%d %H:%M:%S")
    start = (now - timedelta(seconds=60*30)).strftime("%Y-%m-%d %H:%M:%S")

    start = '2022-05-31 12:00:00'
    stop = '2022-05-31 24:00:00'
    wanted_variable_regexes = None


    multipandas2plot = MultiPandas2Plot(larcs,
                                        larc_threads = 1,
                                        plot_threads=None,
                                        plot_vector_items=True,
                                        partition='ATLAS',
                                        start=start,
                                        stop=stop,
                                        run='P1_December13th',
                                        load_pickle='',
                                        save_pickle='',
                                        wanted_variable_regexes=wanted_variable_regexes)

    multipandas2plot.action()



import uhal
import random
import sys
# Creating hardware object to Dymmy Hardware Server UDP though control hub
hw = uhal.getDevice( "dummy" , 'chtcp-2.0://localhost:10203?target=127.0.0.1:50001', f'file://{sys.argv[1]}')

# Creating dictionary with all nodes and respective values read from hardware
d = {}
for node in hw.getNodes():
    d[node] = {'read' : hw.getNode(node).read(), 'written': 0}
hw.dispatch()

# Printing node attributes and writing random value to rw single registers
for key, value in d.items():
    width = f'{hw.getNode(key).getMask():b}'.count('1')
    print(f"{key} = {int(value['read'])}, address: {hw.getNode(key).getAddress():08X}, permission: {hw.getNode(key).getPermission()}, mode: {hw.getNode(key).getMode()}, width: {width}")
    # Check if register is rw, single mode, and a leaf, i.e. empty children nodes
    if hw.getNode(key).getPermission() == uhal.NodePermission(3) and hw.getNode(key).getMode() == uhal.BlockReadWriteMode(0) and not hw.getNode(key).getNodes():
         value['written'] = random.randint(0,2**width-1)
         hw.getNode(key).write(value['written'])
    else:
         value['written'] = 0
hw.dispatch()

# Reading values back and checking against errors
for key, value in d.items():
    value['read'] = hw.getNode(key).read()
hw.dispatch()
for key, value in d.items():
    print(f"{key} => written: {int(value['written'])}, read: {int(value['read'])}")
    # Check if register is rw, single mode, and a leaf, i.e. empty children nodes
    if hw.getNode(key).getPermission() == uhal.NodePermission(3) and hw.getNode(key).getMode() == uhal.BlockReadWriteMode(0) and not hw.getNode(key).getNodes():
        if value['written'] != value['read']:
            raise Exception(f'Value read and written are different for register {key}')
print(f'Simulation using dummy hardware finished with success after checking {len(d)} registers listed in {sys.argv[1]}')


import os
import yaml
from bigtree import nested_dict_to_tree, preorder_iter, print_tree
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, ReadOnly
from cocotb.regression import TestFactory
import random
from operator import attrgetter
import cocotb
import logging

def read_tree(logger):
    yaml_file = os.getenv("BACKANNOTATED_YAML")
    with open(yaml_file, "r") as stream:
        wishlist_dict = yaml.safe_load(stream)
    tree = nested_dict_to_tree(wishlist_dict)
    logger.info('Testing the following register tree:')
    print_tree(tree, all_attrs=True, style='ansi')
    return tree

async def read(dut,address, mask, cycle):
    return await cycle(dut, address, mask, 1, None)


async def write(dut,address, mask, write_values, cycle):
    return await cycle(dut,address, mask, 0, write_values)


async def read_node(dut, node, bus_width, logger, cycle):
    read_values = await read(dut, node.address, node.mask, cycle)
    value = 0
    node_lsb = 0
    logger.debug(f'Reading values from {node.path_name}, read values: {read_values}')
    for i, (addr, msk, rdvl) in enumerate(reversed(list(zip(node.address, node.mask, read_values)))):
        word_width = int(f'{msk:b}'.count('1'))
        word_lsb = f'{{mask:0{bus_width}b}}'.format(mask=msk)[::-1].find('1')
        logger.debug(
            f'Shifting up value (0x{rdvl:x} >> {word_lsb}) from address 0x{addr:x} by {node_lsb} and adding to intermediate sum with value 0x{value:x}. The current word width is {word_width}.')
        value += (rdvl >> word_lsb) << node_lsb
        node_lsb += word_width # incrementing LSB by word width
    return value

def word_mask(width):
    return (1 << width)-1

async def write_node(dut, node, value, bus_width, logger, cycle):
    # Computing the bus mask
    bus_mask = word_mask(bus_width)
    # Reading all the registers associated with the node with the bus mask
    read_values = await read(dut, node.address, [bus_mask]*len(node.address), cycle)
    # Node LSB (can be higher than bus width)
    node_lsb = 0
    # Empty array of values to be written
    write_values = []
    logger.debug(f'Reading values from {node.path_name}, read values: {read_values}')
    for i, (addr, msk, rdvl) in enumerate(reversed(list(zip(node.address, node.mask, read_values)))):
        # Number of bits used by a given node in the current address offset
        word_width = int(f'{msk:b}'.count('1'))
        # Mask to be used to mask the node value
        node_word_mask = word_mask(word_width) << node_lsb
        # Masking the node value
        node_word_value = (value & node_word_mask) >> node_lsb
        # Computing the MSB for current address offset
        lsb = f'{{mask:0{bus_width}b}}'.format(mask=msk)[::-1].find('1')
        logger.debug(f'Word width = {word_width}, node_mask = 0x{node_word_mask:x}, node_word_value = {node_word_value}, lsb = {lsb}')
        # Computing value to be written not yet masked in order to keep current data in the same address offset
        bus_word_value = node_word_value << lsb
        # Masking data that should be kept
        word_to_keep = rdvl & (bus_mask - msk)
        # Computing value to be written and appending to list
        combined = bus_word_value | word_to_keep
        logger.debug(
            f'W:{i} Combining word_to_keep:(0b{{word_to_keep:0{bus_width}b}}, 0x{word_to_keep:x}, {word_to_keep:d}) to bus_word_value: (0b{{bus_word_value:0{bus_width}b}}, 0x{bus_word_value:x}, {bus_word_value:d}), resulting in combined: (0b{{combined:0{bus_width}b}}, 0x{combined:x}, {combined:d})'.format(
                word_to_keep=word_to_keep, bus_word_value=bus_word_value, combined=combined))
        write_values.append(combined)
        # Incrementing node_lsb
        node_lsb += word_width # incrementing LSB by word width
    # Writing combined data back
    logger.debug(f'Writing the following values {write_values[::-1]}')
    ack = await write(dut, node.address, [bus_mask]*len(node.address), write_values[::-1], cycle)
    return True


import json
import xmltodict
import yaml
from bigtree import nested_dict_to_tree, print_tree, preorder_iter, shift_nodes, tree_to_nested_dict
from memory import str2int
from wishlist import HexInt, representer, xml_beautify

default_permission = 'rw'
default_width = 32
default_mask = 2**32-1
maximum_width = 32
propagate_address = True
propagate_mask = True

default_root_attrs = {
    'address_width': 32,
    'address_increment': 4,
    'address_size': 2 ** 16,
    'software_path': '../examples',
    'firmware_path': '../examples',
}

# reading xml to dict string
xml_filename = '../examples/L1CaloGfex.xml'
xml_dict = xmltodict.parse(open(xml_filename, 'rb'))
xml_str = json.dumps(xml_dict)


# manipulating dict string
replacements = {
    '"@address"': '"address"',
    '"@permission"': '"permission"',
    '"@description"': '"description"',
    '"@mask"': '"mask"',
    '"@module"': '"module"',
}
for old, new in replacements.items():
    xml_str = xml_str.replace(old, new)

# generating tree from xml
xml_tree = json.loads(xml_str)['node']
tree = nested_dict_to_tree(xml_tree, name_key='@id', child_key='node')

# copying permission parameter from parents to only tree leaves and assigning default value for leaves without

not_supported_nodes = []
# Iterating though leaves
for leaf in preorder_iter(tree, filter_condition=lambda node: node.is_leaf):
    # copying permission parameter from parents when not defined
    if not hasattr(leaf, 'permission'):
        if hasattr(leaf.parent, 'permission'):
            leaf.permission = leaf.parent.permission
        else:
            # assigning ipbus register permission when parent has no permission defined
            leaf.permission = default_permission
    # copying address parameter from parents when not defined
    if hasattr(leaf, 'address'):
        leaf.address = HexInt(str2int(leaf.address))
    else:
        if hasattr(leaf.parent, 'address'):
            leaf.address = HexInt(str2int(leaf.parent.address))


    # handling mask and computing width if mask is not propagated
    if hasattr(leaf, 'mask'):
        mask = str2int(leaf.mask)
        if not propagate_mask:
            # Computing width based on mask value
            leaf.width = f'{mask:b}'.count('1')
            # deleting mask attribute when defined
            delattr(leaf, 'mask')
        else:
            leaf.mask = HexInt(mask)
    else:
        if propagate_mask:
            leaf.mask = HexInt(default_mask)
        else:
            leaf.width = default_width

    # keeping track of non-supported nodes
    if hasattr(leaf, 'module'): not_supported_nodes.append(leaf.path_name)

# removing non-supported nodes
shift_nodes(tree, not_supported_nodes, [None] * len(not_supported_nodes))

min_address = 2 ** maximum_width - 1
# iterating though all nodes
for node in preorder_iter(tree):
    if hasattr(node, 'address') and not node.is_leaf:
        delattr(node, 'address')
    if hasattr(node, 'address'):
        # finding minimum address to be associated to the root node
        min_address = min(min_address, node.address)
        # removing address when defined, whishlist does not support explicit address yet
        if propagate_address:
            if not node.is_leaf:
                delattr(node, 'address')
        else:
            delattr(node, 'address')
    # deleting permission from branches
    if not node.is_leaf and hasattr(node, 'permission'):
        delattr(node, 'permission')

# Setting minimum address of all nodes as the root address
tree.address = min_address
# Setting default values for
for key, value in default_root_attrs.items():
    setattr(tree, key, value)
print_tree(tree, attr_list=['width', 'length', 'permission', 'mask', 'address'])

# Dumping tree to yaml file
tree_dict = tree_to_nested_dict(tree, all_attrs=True)
with open(xml_filename.replace('xml', 'yaml'), 'w') as file:
    yaml.add_representer(HexInt, representer)
    yaml.dump(tree_dict, file, sort_keys=False)
if not_supported_nodes:
    print(
        f"Warning: Please be aware that the nodes {', '.join(not_supported_nodes)} were removed because ipbus2whislist does not support modules yet.")




import logging
from bigtree import yield_tree, Node, nested_dict_to_tree
import os
import yaml


def registers_to_node(address, mask, read_values, bus_width, logger):
    value = 0
    node_lsb = 0
    for i, (addr, msk, rdvl) in enumerate(reversed(list(zip(address, mask, read_values)))):
        word_width = int(f'{msk:b}'.count('1'))
        word_lsb = f'{{mask:0{bus_width}b}}'.format(mask=msk)[::-1].find('1')
        logger.debug(
            f'Shifting up value (0x{rdvl:x} >> {word_lsb}) from address 0x{addr:x} by {node_lsb} and adding to intermediate sum with value 0x{value:x}. The current word width is {word_width}.')
        value += ((rdvl & msk) >> word_lsb) << node_lsb
        node_lsb += word_width  # incrementing LSB by word width
    return value

def node_to_register(value, address, mask, read_values, bus_width, logger):
    # Computing the bus mask
    bus_mask = word_mask(bus_width)
    # Node LSB (can be higher than bus width)
    node_lsb = 0
    # Empty array of values to be written
    write_values = []
    for i, (addr, msk, rdvl) in enumerate(reversed(list(zip(address, mask, read_values)))):
        # Number of bits used by a given node in the current address offset
        word_width = int(f'{msk:b}'.count('1'))
        # Mask to be used to mask the node value
        node_word_mask = word_mask(word_width) << node_lsb
        # Masking the node value
        node_word_value = (value & node_word_mask) >> node_lsb
        # Computing the MSB for current address offset
        lsb = f'{{mask:0{bus_width}b}}'.format(mask=msk)[::-1].find('1')
        logger.debug(
            f'Word width = {word_width}, node_mask = 0x{node_word_mask:x}, node_word_value = {node_word_value}, lsb = {lsb}')
        # Computing value to be written not yet masked in order to keep current data in the same address offset
        bus_word_value = node_word_value << lsb
        # Masking data that should be kept
        word_to_keep = rdvl & (bus_mask - msk)
        # Computing value to be written and appending to list
        combined = bus_word_value | word_to_keep
        logger.debug(
            f'W:{i} Combining word_to_keep:(0b{{word_to_keep:0{bus_width}b}}, 0x{word_to_keep:x}, {word_to_keep:d}) to bus_word_value: (0b{{bus_word_value:0{bus_width}b}}, 0x{bus_word_value:x}, {bus_word_value:d}), resulting in combined: (0b{{combined:0{bus_width}b}}, 0x{combined:x}, {combined:d})'.format(
                word_to_keep=word_to_keep, bus_word_value=bus_word_value, combined=combined))
        write_values.append(combined)
        # Incrementing node_lsb
        node_lsb += word_width  # incrementing LSB by word width
        # Returning reversed vector, because write values is appended using a for loop in reverse order
        # This way the loop start by the least-significant word. However, the memory mapped allocation places
        # the most significant word first, i.e. at the first address offset of the node
    return write_values[::-1]

def lsb(n):
    return (n & -n).bit_length() - 1
def word_mask(width):
    return (1 << width) - 1


def get_logger(name, level, format=logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')    ):
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setFormatter(format)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger

def read_tree(yaml_file=None, CustomNode=Node):
    if yaml_file is None:
        yaml_file  = os.getenv("BACKANNOTATED_YAML")
    with open(yaml_file, "r") as stream:
        wishlist_dict = yaml.safe_load(stream)
    return nested_dict_to_tree(wishlist_dict, node_type=CustomNode)

def log_tree(tree,logger):
    for branch, stem, node in yield_tree(tree):
        attrs = node.describe(exclude_attributes=["name", 'logger', 'bus_width'], exclude_prefix="_")
        attr_str_list = [f"{k}={v}" for k, v in attrs]
        logger.info(f"{branch}{stem}{node.node_name} {attr_str_list}")


import os
import pandas as pd
pd.options.plotting.backend = "plotly"
from multiprocessing import Pool
from functools import partial
import re
from pathlib import Path
import glob
import logging
logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


# Function outside the class to allow multithreading
def generate_plot(col, run, larc, df, vec_index0_regex, vec_indexany_regex, plot_vector_items):
    def _generate_plot(col, run, larc, df, name,showlegend):
        fig = df[col].plot(template="plotly_white",
                           # title=f'{larc} - {run}'
                           labels=dict(index="Time", value=name,variable='Vector items'))
        # hovermode='x' is used to ensure good browsing performance with large datasets
        fig.update_layout(showlegend=showlegend, hovermode="x")
        fig.write_html(f'../www/{run}/{larc}/{name}.html', include_plotlyjs='directory')

    if 'int' or 'float' in df[col].dtypes.name:
        if pd.options.plotting.backend == 'plotly':
            if vec_index0_regex.search(col):
                name = vec_index0_regex.sub("\\1\\2\\3\\4",col)
                vec_cols = [f'{name}[{index}]' for index in range(int(list(vec_index0_regex.finditer(col))[0].groupdict()['size']))]
                _generate_plot(vec_cols,run,larc,df,name, True)
            if plot_vector_items or not vec_indexany_regex.search(col):
                _generate_plot(col, run, larc, df, col, False)



class Pandas2Plot(object):
    def __init__(self,
                 plot_threads,
                 plot_vector_items,
                 partition='ATLAS',
                 larc='LArC_EMBA_1',
                 start='2022-03-08 16:30:00',
                 stop='2022-03-08 16:40:00',
                 run='current',
                 load_pickle='',
                 save_pickle='is2pandas.pickle',
                 **kwargs):
        self.plot_threads = plot_threads
        self.plot_vector_items = plot_vector_items
        self.vec_index0_regex = re.compile(r"(?P<base>[\.\w]+)(\()(?P<size>\d+)(\))(\(0\))")
        self.vec_indexany_regex = re.compile(r"(?P<base>[\.\w]+)(\()(?P<size>\d+)(\))(\(\d+\))")
        self.link_template = '<a href = "<VARIABLE>.html" target = "main"><VARIABLE></a><br>\n'
        self.link__vector_item_template = '<a href = "<VARIABLE>.html" target = "main"><VARIABLE></a>\n'
        self.link_vector_template = '<button onclick="showhide(\'<VARIABLE>\')" id=<VARIABLE>_b>+</button><br>\n<p hidden id="<VARIABLE>">\n<LINKS_VECTOR_ITEMS></p>'
        self.reading_template('../templates/links_template')
        self.lgr = logging.getLogger(self.__class__.__name__)
        self.lgr.setLevel(logging.DEBUG)
        # Internals
        self.partition = partition
        self.larc = larc
        self.run = run
        self.start = start
        self.stop = stop
        self.load_pickle = load_pickle
        self.save_pickle = save_pickle
        Path(f'../www/{self.run}/{self.larc}/').mkdir(parents=True, exist_ok=True)

    def wanted_variable(self, variable):
        for regex in self.wanted_variable_regexes_compiled:
            if regex.search(variable):
                return True
        return False

    def action(self):
        Path(f'../www').mkdir(parents=True, exist_ok=True)
        Path(f'../pickle').mkdir(parents=True, exist_ok=True)
        self.get_data_from_server()
        df_list = []
        for file in glob.glob('../pickle/*.pickle'):
            df_list.append(pd.read_pickle(file))
        self.df = pd.concat(df_list)
        self.df.sort_index(inplace=True)
        self.generating_plots()
        self.generating_html()

    def get_data_from_server(self):
        os.system(f'rsync --remove-source-files -zvhL zynq-p1.cern.ch:/software/tmp/* ../pickle/')

    def generating_plots(self):
        self.lgr.info(f'Generating plots for {self.larc}')
        gen_plot = partial(generate_plot, run=self.run, larc=self.larc, df=self.df, vec_index0_regex=self.vec_index0_regex,vec_indexany_regex=self.vec_indexany_regex,plot_vector_items=self.plot_vector_items)
        if self.plot_threads == 1:
            for col in self.df.columns:
                gen_plot(col)
        else:
            with Pool() as p:
                p.map(gen_plot, self.df.columns)

    def reading_template(self, name):
        with open(f'{name}.html', 'r') as f:
            setattr(self,name.split('/')[-1],f.read())

    def write_str(self, string, filepath):
        with open(filepath, 'w') as f:
            f.write(string)

    def generating_html(self):
        self.lgr.info(f'Generating pages for {self.larc}.')
        links = ''
        for col in self.df.columns:
            if self.vec_indexany_regex.search(col): # Is it a vector?
                if self.vec_index0_regex.search(col): # Is it the item 0 of a vector?
                    name = self.vec_index0_regex.sub("\\1\\2\\3\\4", col)
                    links += self.link__vector_item_template.replace('<VARIABLE>', name)
                    if self.plot_vector_items: # Should we print the vector sub items?
                        vec_cols = [self.link__vector_item_template.replace('<VARIABLE>', f'{name}[{index}]') for index in
                                    range(int(list(self.vec_index0_regex.finditer(col))[0].groupdict()['size']))]
                        link_vector_items = '<br>'.join(vec_cols)
                        links += self.link_vector_template.replace('<VARIABLE>', name).replace('<LINKS_VECTOR_ITEMS>',link_vector_items)
            else: # Well it is not a vector
                links += self.link_template.replace('<VARIABLE>',col).replace("<TARGET>",'main')
        if not links: links = f'<center>No data from {self.larc} have been found between<br>{self.start} and {self.stop}.<br><br>Either LAr was not running in the ATLAS partition or this LArC was not included.<br>Try selecting another LArC or a different time range.</center>'
        links_html = self.links_template.replace('<LINKS>',links)
        self.write_str(f'<h3>{self.larc}</h3>{links_html}',f'../www/{self.run}/{self.larc}/menu.html')
        os.system(f'cd ../www/{self.run}/{self.larc}; ln -fs ../../perlarc_index.html index.html')


if __name__ == '__main__':
    larc = 'LArC_EMBA_2'
    wanted_variable_regexes = ('bcid_calibration', 'freq', 'counter\[0\]', 'bcr_err', 'period', 'bitslip', 'reset_clock',)
    #wanted_variable_regexes = ('freq', 'counter\[0\]', 'bcr_err', 'period',)
    pandas2plot = Pandas2Plot(larc=larc,run='current', plot_threads=1,plot_vector_items=True, wanted_variable_regexes = wanted_variable_regexes)


# read_words, write_words, and mapper methods are based on code from Emily and Greg
from utils import registers_to_node, node_to_register, get_logger, word_mask
from bigtree import Node
import mmap
import logging
import sys


class wishlist_axi_node(Node):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.value = None
        self.logger = get_logger(self.path_name, logging.INFO)
        self.bus_width = 32

    def mapper(self, fileno, size, offset):
        index = offset % mmap.ALLOCATIONGRANULARITY
        base = offset - index
        length = 4 * size
        maplength = index + length
        mm = mmap.mmap(fileno, maplength, access=mmap.ACCESS_WRITE, offset=base)
        mv = memoryview(mm)
        return mv[index:], length

    def read_words(self):
        with open('/dev/mem', 'r+b') as devmem:
            (mv, length) = self.mapper(devmem.fileno(), len(self.address), self.address[0])
            mv_int = mv.cast('I')
            data = list(mv_int)
            devmem.close()
            return data

    def write_words(self, data):
        with open('/dev/mem', 'w+') as devmem:
            (mv, length) = self.mapper(devmem.fileno(), len(self.address), self.address[0])
            mv_int = mv.cast('I')
            for i, v in enumerate(data):
                mv_int[i] = v
            return True

    def read(self):
        read_values = self.read_words()
        self.logger.debug(f'Reading values from address {self.address}, read values: {read_values}')
        value = registers_to_node(self.address, self.mask, read_values, self.bus_width, self.logger)
        return value

    def write(self, value):
        if not self.permission == 'rw':
            self.logger.critical(f'Terminating application while trying to this node. The respective permission is rw, therefore no value can not be written to it!' )
            sys.exit()
        # Reading all the registers associated with the node with the bus mask if any mask bit is 0
        if self.mask != [word_mask(self.bus_width) for _ in range(len(self.mask))]:
            read_values = self.read_words()
        else:
            read_values = [0 for _ in range(len(self.mask))]
        # Writing combined data back
        write_values = node_to_register(value, self.address, self.mask, read_values, self.bus_width, self.logger)
        self.logger.debug(f'Writing the following values {write_values}')
        self.write_words(write_values)
        return True

    def convert(self, value, parameter, **kwargs):
        if hasattr(self, parameter):
            if value == (1 << self.width) -1:
                self.logger.warning('Attempted conversion returned -1 because read value is saturated (reached maximum value due overflow protection)')
                return -1
            else:
                return eval(getattr(self, parameter))
        else:
            return value



import os
import yaml
from bigtree import nested_dict_to_tree, preorder_iter, print_tree
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, ReadOnly
from cocotb.regression import TestFactory
import random
from operator import attrgetter
import cocotb
import logging
from node import read_tree, read_node, write_node


async def cycle(dut, address, mask, read_mode, write_values):
    read_values = []
    # Iterating for one clock cycle more than the number of words in the node because address decoder outputs is registered
    # Also delaying mask for one iteration because the mask is needed in the iteration i+1 
    for i, (addr, msk) in enumerate(zip(address+[None],[None]+mask)):
        await RisingEdge(dut.clk_i)
        # Write Stage of the clock cycle 
        if i < len(address):
           dut.read_i.value = bool(read_mode)
           dut.write_i.value = not(bool(read_mode))
           dut.address_i.value = addr
           if not read_mode:
               dut.data_i.value = write_values[i]
        else:
           dut.read_i.value = 0
           dut.write_i.value = 0
        # Read stage of the clock cycle
        if read_mode:
            await ReadOnly()
            if i > 0:
                read_values.append(dut.data_o.value.integer & msk)
    if read_mode:
        return read_values
    else:
        return True


@cocotb.coroutine
async def register_test(dut, logger, tree, shufle_order=1):
    """Testing registers"""
    # Configuring clock
    clock = Clock(dut.clk_i, 10, units="ns")  # Create a 10ns period clock on port clk
    cocotb.start_soon(clock.start(start_high=False)) # Start the clock. Start it low to avoid issues on the first RisingEdge

    # Initializing contr78ol signals
    dut.read_i.value = 0
    dut.write_i.value = 0
    dut.data_i.value = 0
    dut.address_i.value = 0
    bus_width = len(dut.data_i)

    # Extracting tree of nodes
    nodes = list(preorder_iter(tree, filter_condition=lambda node: node.is_leaf))

    # Writing stimullus
    if shufle_order: random.shuffle(nodes)
    for node in nodes:
        logger.info(f'Writing stimulus for {node.path_name}')
        node.stimulus = random.randint(0,2**node.width-1)
        path = node.path_name.lower().split('/')
        if node.permission == 'r':
            signal = attrgetter(f"{path[1]}_status_i.{'.'.join(path[2:])}")(dut)
            signal.value = node.stimulus
        else:
            ack = await write_node(dut,node,node.stimulus, bus_width, logger, cycle)

    # Checking stimulus
    if shufle_order: random.shuffle(nodes)
    for node in nodes:
        logger.info(f'Checking node: {node.path_name}, permission: {node.permission}')
        node_value = await read_node(dut, node, bus_width, logger, cycle)
        logger.debug(f'Stimulus = {node.stimulus}, actual= {node_value}')
        assert node.stimulus == node_value, f'Actual data for Node {node.path_name} {node_value} is different than applied stimulus {node.stimulus}'




#if __name__ == '__main__':
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
tree = read_tree(logger)
# Factory of tests
factory = TestFactory(register_test)
factory.add_option("shufle_order", [False, True, True, True, True])
factory.add_option("logger", [logger])
factory.add_option("tree", [tree])
factory.generate_tests()






import os
import yaml
from bigtree import nested_dict_to_tree, preorder_iter, print_tree
from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, ReadOnly
from cocotb.regression import TestFactory
import random
from operator import attrgetter
import cocotb
import logging
from cocotbext.axi import AxiBus, AxiLiteBus, AxiMaster, AxiLiteMaster, AxiLiteSlave, AxiSlave, AxiResp
from rtl_simulation import read_tree, write_node, read_node


async def cycle(axilite_master, address, mask, read_mode, write_values):
    read_values = []
    # Iterating for one clock cycle more than the number of words in the node because address decoder outputs is registered
    # Also delaying mask for one iteration because the mask is needed in the iteration i+1
    for i, (addr, msk) in enumerate(zip(address, mask)):
        # Write Stage of the clock cycle
        if not read_mode:
            bytes = int(write_values[i]).to_bytes(4, byteorder='little')
            (address_resp, length, resp) = await axilite_master.write(addr, bytes)
        else:
            (address_resp, data, resp) = await axilite_master.read(addr, 4)
            integer = int.from_bytes(data, byteorder='little', signed=False)
            read_values.append(integer & msk)
        if resp != AxiResp.OKAY:
            raise Exception(f'AXI transfer with read_mode = {read_mode}, addr= {addr} returned response code {resp}')
    if read_mode:
        return read_values
    else:
        return True


@cocotb.coroutine
async def register_test(dut, logger, tree, shufle_order=1):
    """Testing registers"""
    # Configuring clock
    cocotb.start_soon(Clock(dut.S_AXI_ACLK, 2, units="ns").start())
    axilite_master = AxiLiteMaster(AxiLiteBus.from_prefix(dut, "S_AXI"), dut.S_AXI_ACLK, dut.S_AXI_ARESETN,
                                   reset_active_level=False)
    bus_width = len(dut.Bus2IP_Data)
    await cycle_reset(dut.S_AXI_ACLK, dut.S_AXI_ARESETN)

    # register transfers
    # Extracting tree of nodes
    nodes = list(preorder_iter(tree, filter_condition=lambda node: node.is_leaf))

    # Writing stimullus
    if shufle_order: random.shuffle(nodes)
    for node in nodes:
        logger.info(f'Writing stimulus for {node.path_name}')
        path = node.path_name.lower().split('/')
        if node.permission == 'rw':
            node.stimulus = random.randint(0, 2 ** node.width - 1)
            ack = await write_node(axilite_master, node, node.stimulus, bus_width, logger, cycle)

    # Checking stimulus
    if shufle_order: random.shuffle(nodes)
    for node in nodes:
        logger.info(f'Checking node: {node.path_name}, permission: {node.permission}')
        node_value = await read_node(axilite_master, node, bus_width, logger, cycle)
        logger.debug(f'Stimulus = {node.stimulus}, actual= {node_value}')
        assert node.stimulus == node_value, f'Actual data for Node {node.path_name} {node_value} is different than applied stimulus {node.stimulus}'


async def cycle_reset(clk, rst):
    rst.setimmediatevalue(1)
    await RisingEdge(clk)
    await RisingEdge(clk)
    rst.value = 0
    await RisingEdge(clk)
    await RisingEdge(clk)
    rst.value = 1
    await RisingEdge(clk)
    await RisingEdge(clk)


# if __name__ == '__main__':
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
tree = read_tree(logger)
# Factory of tests
factory = TestFactory(register_test)
factory.add_option("shufle_order", [False, True, True, True, True])
factory.add_option("logger", [logger])
factory.add_option("tree", [tree])
factory.generate_tests()



import os
from itertools import cycle
import pandas as pd
from memory import inclusive_range

def index_string_difference(s1,s2):
    # s1 should be the smaller string
    diffs = [i for i,(a,b)  in enumerate(zip(cycle([*s1]),[*s2])) if a != b]
    if any(diffs):
        return diffs[0]
    else:
        return 0

def diff_string(common, register):
    return register[index_string_difference(common,register):]

def formatting(space, space_style, wishlist_dict, sep='.'):
    # Finding common names
    for index, row in space.iterrows():
        # removing unused bits, removing first / and replacing remaining with desited separator
        row = row.dropna().str.replace('/','',n=1).str.replace('/',sep)
        if any(row):
            common = os.path.commonprefix(list(row.values)).split('(')[0]
            space.loc[index,'Common name'] = common
            for bit in row.index:
                space.loc[index,bit] = diff_string(common, row[bit])
    # Adding hexadecimal address offset
    space['Address'] = space.index.map(lambda x: f"0x{{0:0{wishlist_dict['address_width']//4}X}}".format(x))
    # Changing the order of the columns
    space = space[['Address', 'Common name'] + list(inclusive_range(wishlist_dict['address_width']-1,0,-1))]
    space_style['Address'] = 'border: 1px solid black; background-color: AliceBlue'
    space_style['Common name'] = 'border: 1px solid black; background-color: FloralWhite'
    return space, space_style










